import { OrderedSet } from "immutable";
import { Type, NamedType, ClassType, EnumType, UnionType } from "./Type";
import { Namespace, Name, Namer, DependencyName } from "./Naming";
import { Renderer, BlankLineLocations } from "./Renderer";
import { Sourcelike } from "./Source";
export declare abstract class ConvenienceRenderer extends Renderer {
    protected globalNamespace: Namespace;
    private _topLevelNames;
    private _namesForNamedTypes;
    private _propertyNames;
    private _caseNames;
    private _namedTypes;
    private _namedClasses;
    private _namedEnums;
    private _namedUnions;
    private _haveUnions;
    protected readonly forbiddenNamesForGlobalNamespace: string[];
    protected forbiddenForProperties(c: ClassType, classNamed: Name): {
        names: Name[];
        namespaces: Namespace[];
    };
    protected forbiddenForCases(e: EnumType, enumNamed: Name): {
        names: Name[];
        namespaces: Namespace[];
    };
    protected topLevelDependencyNames(topLevelName: Name): DependencyName[];
    protected namedTypeDependencyNames(t: NamedType, name: Name): DependencyName[];
    protected abstract topLevelNameStyle(rawName: string): string;
    protected readonly abstract namedTypeNamer: Namer;
    protected readonly abstract propertyNamer: Namer;
    protected readonly abstract caseNamer: Namer;
    protected abstract namedTypeToNameForTopLevel(type: Type): NamedType | null;
    protected abstract emitSourceStructure(): void;
    protected readonly casesInGlobalNamespace: boolean;
    protected unionNeedsName(u: UnionType): boolean;
    protected setUpNaming(): Namespace[];
    private nameForTopLevel;
    private addNamedForNamedType;
    private addPropertyNameds;
    private addCaseNameds;
    private childrenOfType;
    protected readonly namedUnions: OrderedSet<UnionType>;
    protected readonly haveNamedUnions: boolean;
    protected readonly haveUnions: boolean;
    protected readonly enums: OrderedSet<EnumType>;
    protected readonly haveEnums: boolean;
    protected unionFieldName: (fieldType: Type) => string;
    protected nameForNamedType: (t: NamedType) => Name;
    protected forEachTopLevel: (blankLocations: BlankLineLocations, f: (t: Type, name: Name) => void, predicate?: ((t: Type) => boolean) | undefined) => void;
    protected forEachProperty: (c: ClassType, blankLocations: BlankLineLocations, f: (name: Name, jsonName: string, t: Type) => void) => void;
    protected forEachCase: (e: EnumType, blankLocations: BlankLineLocations, f: (name: Name, jsonName: string) => void) => void;
    protected callForNamedType<T extends NamedType>(t: T, f: (t: T, name: Name) => void): void;
    protected forEachSpecificNamedType<T extends NamedType>(blankLocations: BlankLineLocations, types: OrderedSet<T>, f: (t: T, name: Name) => void): void;
    protected forEachClass: (blankLocations: BlankLineLocations, f: (c: ClassType, className: Name) => void) => void;
    protected forEachEnum: (blankLocations: BlankLineLocations, f: (u: EnumType, enumName: Name) => void) => void;
    protected forEachUnion: (blankLocations: BlankLineLocations, f: (u: UnionType, unionName: Name) => void) => void;
    protected forEachUniqueUnion<T>(blankLocations: BlankLineLocations, uniqueValue: (u: UnionType) => T, f: (firstUnion: UnionType, value: T) => void): void;
    protected forEachNamedType: (blankLocations: BlankLineLocations, leavesFirst: boolean, classFunc: (c: ClassType, className: Name) => void, enumFunc: (e: EnumType, enumName: Name) => void, unionFunc: (u: UnionType, unionName: Name) => void) => void;
    protected sourcelikeToString: (src: Sourcelike) => string;
    protected emitSource(): void;
}
