"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Naming_1 = require("./Naming");
var Renderer_1 = require("./Renderer");
var Support_1 = require("./Support");
var Source_1 = require("./Source");
var ConvenienceRenderer = /** @class */ (function (_super) {
    __extends(ConvenienceRenderer, _super);
    function ConvenienceRenderer() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.nameForTopLevel = function (type, name) {
            var maybeNamedType = _this.namedTypeToNameForTopLevel(type);
            var styledName;
            if (maybeNamedType) {
                styledName = _this.namedTypeNamer.nameStyle(name);
            }
            else {
                styledName = _this.topLevelNameStyle(name);
            }
            var named = _this.globalNamespace.add(new Naming_1.FixedName(styledName));
            var dependencyNames = _this.topLevelDependencyNames(named);
            for (var _i = 0, dependencyNames_1 = dependencyNames; _i < dependencyNames_1.length; _i++) {
                var dn = dependencyNames_1[_i];
                _this.globalNamespace.add(dn);
            }
            if (maybeNamedType) {
                _this._namesForNamedTypes = _this._namesForNamedTypes.set(maybeNamedType, named);
            }
            return named;
        };
        _this.addNamedForNamedType = function (type) {
            var existing = _this._namesForNamedTypes.get(type);
            if (existing !== undefined)
                return existing;
            var name = type.names.combined;
            var named = _this.globalNamespace.add(new Naming_1.SimpleName(name, _this.namedTypeNamer));
            var dependencyNames = _this.namedTypeDependencyNames(type, named);
            for (var _i = 0, dependencyNames_2 = dependencyNames; _i < dependencyNames_2.length; _i++) {
                var dn = dependencyNames_2[_i];
                _this.globalNamespace.add(dn);
            }
            _this._namesForNamedTypes = _this._namesForNamedTypes.set(type, named);
            return named;
        };
        _this.addPropertyNameds = function (c, classNamed) {
            var _a = _this.forbiddenForProperties(c, classNamed), forbiddenNames = _a.names, forbiddenNamespace = _a.namespaces;
            var ns = new Naming_1.Namespace(c.names.combined, _this.globalNamespace, immutable_1.Set(forbiddenNamespace), immutable_1.Set(forbiddenNames));
            var names = c.properties
                .map(function (t, name) {
                return ns.add(new Naming_1.SimpleName(name, _this.propertyNamer));
            })
                .toMap();
            _this._propertyNames = _this._propertyNames.set(c, names);
        };
        // FIXME: this is very similar to addPropertyNameds
        _this.addCaseNameds = function (e, enumNamed) {
            var _a = _this.forbiddenForCases(e, enumNamed), forbiddenNames = _a.names, forbiddenNamespace = _a.namespaces;
            var ns;
            if (_this.casesInGlobalNamespace) {
                ns = _this.globalNamespace;
            }
            else {
                ns = new Naming_1.Namespace(e.names.combined, _this.globalNamespace, immutable_1.Set(forbiddenNamespace), immutable_1.Set(forbiddenNames));
            }
            var names = immutable_1.Map();
            e.cases.forEach(function (name) {
                names = names.set(name, ns.add(new Naming_1.SimpleName(name, _this.caseNamer)));
            });
            _this._caseNames = _this._caseNames.set(e, names);
        };
        _this.childrenOfType = function (t) {
            var names = _this.names;
            if (t instanceof Type_1.ClassType) {
                var propertyNameds_1 = Support_1.defined(_this._propertyNames.get(t));
                return t.properties
                    .sortBy(function (_, n) { return Support_1.defined(names.get(Support_1.defined(propertyNameds_1.get(n)))); })
                    .toOrderedSet();
            }
            return t.children.toOrderedSet();
        };
        // FIXME: These should use Name.  Not only to fix collisions, but also
        // so we can rename them via config at some point.
        _this.unionFieldName = function (fieldType) {
            var typeNameForUnionMember = function (t) {
                return Type_1.matchType(t, function (anyType) { return "anything"; }, function (nullType) { return "null"; }, function (boolType) { return "bool"; }, function (integerType) { return "integer"; }, function (doubleType) { return "double"; }, function (stringType) { return "string"; }, function (arrayType) { return typeNameForUnionMember(arrayType.items) + "_array"; }, function (classType) { return Support_1.defined(_this.names.get(_this.nameForNamedType(classType))); }, function (mapType) { return typeNameForUnionMember(mapType.values) + "_map"; }, function (enumType) { return "enum"; }, function (unionType) { return "union"; });
            };
            return _this.propertyNamer.nameStyle(typeNameForUnionMember(fieldType));
        };
        _this.nameForNamedType = function (t) {
            if (!_this._namesForNamedTypes.has(t)) {
                throw "Named type does not exist.";
            }
            return Support_1.defined(_this._namesForNamedTypes.get(t));
        };
        _this.forEachTopLevel = function (blankLocations, f, predicate) {
            var topLevels;
            if (predicate) {
                topLevels = _this.topLevels.filter(predicate);
            }
            else {
                topLevels = _this.topLevels;
            }
            _this.forEachWithBlankLines(topLevels, blankLocations, function (t, name) {
                return f(t, Support_1.defined(_this._topLevelNames.get(name)));
            });
        };
        _this.forEachProperty = function (c, blankLocations, f) {
            var propertyNames = Support_1.defined(_this._propertyNames.get(c));
            var sortedPropertyNames = propertyNames.sortBy(function (n) { return _this.names.get(n); }).toOrderedMap();
            _this.forEachWithBlankLines(sortedPropertyNames, blankLocations, function (name, jsonName) {
                var t = Support_1.defined(c.properties.get(jsonName));
                f(name, jsonName, t);
            });
        };
        _this.forEachCase = function (e, blankLocations, f) {
            var caseNames = Support_1.defined(_this._caseNames.get(e));
            var sortedCaseNames = caseNames.sortBy(function (n) { return _this.names.get(n); }).toOrderedMap();
            _this.forEachWithBlankLines(sortedCaseNames, blankLocations, f);
        };
        _this.forEachClass = function (blankLocations, f) {
            _this.forEachSpecificNamedType(blankLocations, _this._namedClasses, f);
        };
        _this.forEachEnum = function (blankLocations, f) {
            _this.forEachSpecificNamedType(blankLocations, _this._namedEnums, f);
        };
        _this.forEachUnion = function (blankLocations, f) {
            _this.forEachSpecificNamedType(blankLocations, _this._namedUnions, f);
        };
        _this.forEachNamedType = function (blankLocations, leavesFirst, classFunc, enumFunc, unionFunc) {
            var collection = _this._namedTypes;
            if (leavesFirst)
                collection = collection.reverse();
            _this.forEachWithBlankLines(collection, blankLocations, function (t) {
                if (t instanceof Type_1.ClassType) {
                    _this.callForNamedType(t, classFunc);
                }
                else if (t instanceof Type_1.EnumType) {
                    _this.callForNamedType(t, enumFunc);
                }
                else if (t instanceof Type_1.UnionType) {
                    _this.callForNamedType(t, unionFunc);
                }
                else {
                    throw "Named type that's neither a class nor union";
                }
            });
        };
        // You should never have to use this to produce parts of your generated
        // code.  If you need to modify a Name, for example to change its casing,
        // use `modifySource`.
        _this.sourcelikeToString = function (src) {
            return Source_1.serializeRenderResult({
                rootSource: Source_1.sourcelikeToSource(src),
                names: _this.names
            }, 
            // FIXME: Use proper indentation.
            "").lines.join("\n");
        };
        return _this;
    }
    Object.defineProperty(ConvenienceRenderer.prototype, "forbiddenNamesForGlobalNamespace", {
        get: function () {
            return [];
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.forbiddenForProperties = function (c, classNamed) {
        return { names: [], namespaces: [] };
    };
    ConvenienceRenderer.prototype.forbiddenForCases = function (e, enumNamed) {
        return { names: [], namespaces: [] };
    };
    ConvenienceRenderer.prototype.topLevelDependencyNames = function (topLevelName) {
        return [];
    };
    ConvenienceRenderer.prototype.namedTypeDependencyNames = function (t, name) {
        return [];
    };
    Object.defineProperty(ConvenienceRenderer.prototype, "casesInGlobalNamespace", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.unionNeedsName = function (u) {
        return !Type_1.nullableFromUnion(u);
    };
    ConvenienceRenderer.prototype.setUpNaming = function () {
        var _this = this;
        this.globalNamespace = Naming_1.keywordNamespace("global", this.forbiddenNamesForGlobalNamespace);
        var _a = Type_1.allNamedTypesSeparated(this.topLevels), classes = _a.classes, enums = _a.enums, unions = _a.unions;
        var namedUnions = unions.filter(function (u) { return _this.unionNeedsName(u); }).toOrderedSet();
        this._namesForNamedTypes = immutable_1.Map();
        this._propertyNames = immutable_1.Map();
        this._caseNames = immutable_1.Map();
        this._topLevelNames = this.topLevels.map(this.nameForTopLevel).toMap();
        classes.forEach(function (c) {
            var named = _this.addNamedForNamedType(c);
            _this.addPropertyNameds(c, named);
        });
        namedUnions.forEach(function (u) { return _this.addNamedForNamedType(u); });
        enums.forEach(function (e) {
            var named = _this.addNamedForNamedType(e);
            _this.addCaseNameds(e, named);
        });
        return [this.globalNamespace];
    };
    Object.defineProperty(ConvenienceRenderer.prototype, "namedUnions", {
        get: function () {
            return this._namedUnions;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "haveNamedUnions", {
        get: function () {
            return !this._namedUnions.isEmpty();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "haveUnions", {
        get: function () {
            return this._haveUnions;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "enums", {
        get: function () {
            return this._namedEnums;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "haveEnums", {
        get: function () {
            return !this._namedEnums.isEmpty();
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.callForNamedType = function (t, f) {
        f(t, Support_1.defined(this._namesForNamedTypes.get(t)));
    };
    ConvenienceRenderer.prototype.forEachSpecificNamedType = function (blankLocations, types, f) {
        var _this = this;
        this.forEachWithBlankLines(types, blankLocations, function (t) {
            _this.callForNamedType(t, f);
        });
    };
    ConvenienceRenderer.prototype.forEachUniqueUnion = function (blankLocations, uniqueValue, f) {
        var firstUnionByValue = immutable_1.OrderedMap();
        this._namedUnions.forEach(function (u) {
            var v = uniqueValue(u);
            if (!firstUnionByValue.has(v)) {
                firstUnionByValue = firstUnionByValue.set(v, u);
            }
        });
        this.forEachWithBlankLines(firstUnionByValue, blankLocations, f);
    };
    ConvenienceRenderer.prototype.emitSource = function () {
        var _this = this;
        var types = Type_1.allNamedTypes(this.topLevels, this.childrenOfType);
        this._haveUnions = types.some(function (t) { return t instanceof Type_1.UnionType; });
        this._namedTypes = types
            .filter(function (t) { return !(t instanceof Type_1.UnionType) || _this.unionNeedsName(t); })
            .toOrderedSet();
        var _a = Type_1.separateNamedTypes(this._namedTypes), classes = _a.classes, enums = _a.enums, unions = _a.unions;
        this._namedClasses = classes;
        this._namedEnums = enums;
        this._namedUnions = unions;
        this.emitSourceStructure();
    };
    return ConvenienceRenderer;
}(Renderer_1.Renderer));
exports.ConvenienceRenderer = ConvenienceRenderer;
