"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
function glueTypeNamesToNative(_a) {
    var names = _a.names, combined = _a.combined;
    return { names: immutable_1.Set(names), combined: combined };
}
function glueTypeToNative(type, classes) {
    switch (type.kind) {
        case "array": {
            var items = glueTypeToNative(type.items, classes);
            return new Type_1.ArrayType(items);
        }
        case "class": {
            var c = classes[type.index];
            if (c === null) {
                throw "Expected class is not in graph array";
            }
            return c;
        }
        case "map": {
            var values = glueTypeToNative(type.values, classes);
            return new Type_1.MapType(values);
        }
        case "enum": {
            return new Type_1.EnumType(glueTypeNamesToNative(type.names), immutable_1.OrderedSet(type.cases));
        }
        case "union": {
            var members = type.members.map(function (t) { return glueTypeToNative(t, classes); });
            return new Type_1.UnionType(glueTypeNamesToNative(type.names), immutable_1.OrderedSet(members));
        }
        default:
            return new Type_1.PrimitiveType(type.kind);
    }
}
function glueTypesToNative(glueEntries) {
    var classes = [];
    for (var _i = 0, glueEntries_1 = glueEntries; _i < glueEntries_1.length; _i++) {
        var c = glueEntries_1[_i];
        if (c === null) {
            classes.push(null);
        }
        else {
            classes.push(new Type_1.ClassType(glueTypeNamesToNative(c.names)));
        }
    }
    for (var i = 0; i < classes.length; i++) {
        var c = classes[i];
        if (c === null) {
            continue;
        }
        var glueProperties = immutable_1.Map(glueEntries[i].properties);
        c.setProperties(glueProperties.map(function (t) { return glueTypeToNative(t, classes); }).toMap());
    }
    return classes;
}
function glueGraphToNative(glueGraph) {
    var classes = glueTypesToNative(glueGraph.classes);
    var result = immutable_1.Map(glueGraph.toplevels)
        .map(function (t) { return glueTypeToNative(t, classes); })
        .toMap();
    return result;
}
exports.glueGraphToNative = glueGraphToNative;
