"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var lodash_1 = require("lodash");
var TargetLanguage_1 = require("../TargetLanguage");
var RendererOptions_1 = require("../RendererOptions");
var purescript_1 = require("../purescript");
var Support_1 = require("../Support");
var Renderers = require("../Language.Renderers")
var CSharp_1 = require("./CSharp");
var Golang_1 = require("./Golang");
var CPlusPlus_1 = require("./CPlusPlus");
var Java_1 = require("./Java");
var SimpleTypes_1 = require("./SimpleTypes");
var TypeScript_1 = require("./TypeScript");
var Swift_1 = require("./Swift");
var SwiftVersion;
(function (SwiftVersion) {
    SwiftVersion[SwiftVersion["Swift3"] = 0] = "Swift3";
    SwiftVersion[SwiftVersion["Swift4"] = 1] = "Swift4";
})(SwiftVersion || (SwiftVersion = {}));
var SwiftTargetLanguage = /** @class */ (function (_super) {
    __extends(SwiftTargetLanguage, _super);
    function SwiftTargetLanguage() {
        var _this = this;
        var versionOption = new RendererOptions_1.EnumOption("swift-version", "Swift version", [
            ["4", SwiftVersion.Swift4],
            ["3", SwiftVersion.Swift3]
        ]);
        var swift4 = new Swift_1.default();
        var optionDefinitions = [versionOption.definition].concat(swift4.optionDefinitions);
        _this = _super.call(this, "Swift", ["swift"], "swift", optionDefinitions) || this;
        _this._versionOption = versionOption;
        _this._swift3 = new TargetLanguage_1.PureScriptTargetLanguage(purescript_1.fromJust(Renderers.rendererForLanguage("swift3")));
        _this._swift4 = swift4;
        return _this;
    }
    SwiftTargetLanguage.prototype.transformAndRenderConfig = function (config) {
        var version = this._versionOption.getValue(config.rendererOptions);
        switch (version) {
            case SwiftVersion.Swift4:
                return this._swift4.transformAndRenderConfig(config);
            case SwiftVersion.Swift3:
                return this._swift3.transformAndRenderConfig(config);
            default:
                return Support_1.assertNever(version);
        }
    };
    return SwiftTargetLanguage;
}(TargetLanguage_1.TargetLanguage));
var typeScriptTargetLanguages = [
    new CSharp_1.default(),
    new Golang_1.default(),
    new CPlusPlus_1.default(),
    new Java_1.default(),
    new TypeScript_1.default(),
    new SwiftTargetLanguage(),
    new SimpleTypes_1.default()
];
var pureScriptTargetLanguages = Renderers.all
    .filter(function (r) { return r.names[0] !== "swift3"; })
    .map(function (r) { return new TargetLanguage_1.PureScriptTargetLanguage(r); });
exports.all = typeScriptTargetLanguages.concat(pureScriptTargetLanguages);
function languageNamed(name) {
    return lodash_1.find(exports.all, function (l) { return lodash_1.includes(l.names, name) || l.displayName === name; });
}
exports.languageNamed = languageNamed;
