"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var TargetLanguage_1 = require("../TargetLanguage");
var Type_1 = require("../Type");
var Naming_1 = require("../Naming");
var RendererOptions_1 = require("../RendererOptions");
var Source_1 = require("../Source");
var Annotation_1 = require("../Annotation");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var Support_1 = require("../Support");
var Swift4TargetLanguage = /** @class */ (function (_super) {
    __extends(Swift4TargetLanguage, _super);
    function Swift4TargetLanguage() {
        var _this = this;
        var justTypesOption = new RendererOptions_1.BooleanOption("just-types", "Plain types only", false);
        var classOption = new RendererOptions_1.EnumOption("struct-or-class", "Generate structs or classes", [
            ["struct", false],
            ["class", true]
        ]);
        var options = [justTypesOption, classOption];
        _this = _super.call(this, "Swift 4", ["swift4"], "swift", options.map(function (o) { return o.definition; })) || this;
        _this._justTypesOption = justTypesOption;
        _this._classOption = classOption;
        return _this;
    }
    Swift4TargetLanguage.prototype.renderGraph = function (topLevels, optionValues) {
        var renderer = new SwiftRenderer(topLevels, this._justTypesOption.getValue(optionValues), this._classOption.getValue(optionValues));
        return renderer.render();
    };
    return Swift4TargetLanguage;
}(TargetLanguage_1.TypeScriptTargetLanguage));
exports.default = Swift4TargetLanguage;
var keywords = [
    "associatedtype",
    "class",
    "deinit",
    "enum",
    "extension",
    "fileprivate",
    "func",
    "import",
    "init",
    "inout",
    "internal",
    "let",
    "open",
    "operator",
    "private",
    "protocol",
    "public",
    "static",
    "struct",
    "subscript",
    "typealias",
    "var",
    "break",
    "case",
    "continue",
    "default",
    "defer",
    "do",
    "else",
    "fallthrough",
    "for",
    "guard",
    "if",
    "in",
    "repeat",
    "return",
    "switch",
    "where",
    "while",
    "as",
    "Any",
    "catch",
    "false",
    "is",
    "nil",
    "rethrows",
    "super",
    "self",
    "Self",
    "throw",
    "throws",
    "true",
    "try",
    "_",
    "associativity",
    "convenience",
    "dynamic",
    "didSet",
    "final",
    "get",
    "infix",
    "indirect",
    "lazy",
    "left",
    "mutating",
    "nonmutating",
    "optional",
    "override",
    "postfix",
    "precedence",
    "prefix",
    "Protocol",
    "required",
    "right",
    "set",
    "Type",
    "unowned",
    "weak",
    "willSet",
    "String",
    "Int",
    "Double",
    "Bool",
    "Data",
    "CommandLine",
    "FileHandle",
    "JSONSerialization",
    "checkNull",
    "removeNSNull",
    "nilToNSNull",
    "convertArray",
    "convertOptional",
    "convertDict",
    "convertDouble"
];
function isPartCharacter(codePoint) {
    return Support_1.isLetterOrUnderscore(codePoint) || Support_1.isNumeric(codePoint);
}
function isStartCharacter(codePoint) {
    return isPartCharacter(codePoint) && !Support_1.isDigit(codePoint);
}
var legalizeName = Support_1.legalizeCharacters(isPartCharacter);
function swiftNameStyle(isUpper, original) {
    var legalized = legalizeName(original);
    var cameled = Support_1.pascalCase(legalized);
    return Support_1.startWithLetter(isStartCharacter, isUpper, cameled);
}
function unicodeEscape(codePoint) {
    return "\\u{" + Support_1.intToHex(codePoint, 0) + "}";
}
var stringEscape = Support_1.utf32ConcatMap(Support_1.escapeNonPrintableMapper(Support_1.isPrintable, unicodeEscape));
var upperNamingFunction = Naming_1.funPrefixNamer(function (s) { return swiftNameStyle(true, s); });
var lowerNamingFunction = Naming_1.funPrefixNamer(function (s) { return swiftNameStyle(false, s); });
var SwiftRenderer = /** @class */ (function (_super) {
    __extends(SwiftRenderer, _super);
    function SwiftRenderer(topLevels, _justTypes, _useClasses) {
        var _this = _super.call(this, topLevels) || this;
        _this._justTypes = _justTypes;
        _this._useClasses = _useClasses;
        _this.emitBlock = function (line, f) {
            _this.emitLine(line, " {");
            _this.indent(f);
            _this.emitLine("}");
        };
        _this.swift3OrPlainCase = function (swift3OrPlain, swift4NonPlain) {
            if (_this._justTypes)
                return swift3OrPlain;
            else
                return swift4NonPlain;
        };
        _this.swiftType = function (t, withIssues) {
            if (withIssues === void 0) { withIssues = false; }
            return Type_1.matchType(t, function (anyType) { return Source_1.maybeAnnotated(withIssues, Annotation_1.anyTypeIssueAnnotation, _this.swift3OrPlainCase("Any?", "JSONAny")); }, function (nullType) {
                return Source_1.maybeAnnotated(withIssues, Annotation_1.nullTypeIssueAnnotation, _this.swift3OrPlainCase("NSNull", "JSONNull?"));
            }, function (boolType) { return "Bool"; }, function (integerType) { return "Int"; }, function (doubleType) { return "Double"; }, function (stringType) { return "String"; }, function (arrayType) { return ["[", _this.swiftType(arrayType.items, withIssues), "]"]; }, function (classType) { return _this.nameForNamedType(classType); }, function (mapType) { return ["[String: ", _this.swiftType(mapType.values, withIssues), "]"]; }, function (enumType) { return _this.nameForNamedType(enumType); }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable)
                    return [_this.swiftType(nullable, withIssues), "?"];
                return _this.nameForNamedType(unionType);
            });
        };
        _this.unionFieldName = function (fieldType) {
            return Type_1.matchType(fieldType, function (anyType) { return "anything"; }, function (nullType) { return "null"; }, function (boolType) { return "bool"; }, function (integerType) { return "integer"; }, function (doubleType) { return "double"; }, function (stringType) { return "string"; }, function (arrayType) { return _this.unionFieldName(arrayType.items) + "Array"; }, function (classType) { return Support_1.decapitalize(Support_1.defined(_this.names.get(_this.nameForNamedType(classType)))); }, function (mapType) { return _this.unionFieldName(mapType.values) + "Map"; }, function (enumType) { return "enumeration"; }, function (unionType) { return "oneOf"; });
        };
        _this.renderHeader = function () {
            if (!_this._justTypes) {
                _this.emitLine("// To parse the JSON, add this file to your project and do:");
                _this.emitLine("//");
                _this.forEachTopLevel("none", function (t, name) {
                    _this.emitLine("//   let ", Source_1.modifySource(Support_1.camelCase, name), " = ", name, ".from(json: jsonString)!");
                });
                _this.emitNewline();
            }
            _this.emitLine("import Foundation");
        };
        _this.renderTopLevelAlias = function (t, name) {
            _this.emitLine("typealias ", name, " = ", _this.swiftType(t, true));
        };
        _this.getCodableString = function () {
            return _this.swift3OrPlainCase("", ": Codable");
        };
        _this.renderClassDefinition = function (c, className) {
            var structOrClass = _this._useClasses ? "class" : "struct";
            var codableString = _this.getCodableString();
            _this.emitBlock([structOrClass, " ", className, codableString], function () {
                _this.forEachProperty(c, "none", function (name, jsonName, t) {
                    _this.emitLine("let ", name, ": ", _this.swiftType(t, true));
                });
            });
        };
        _this.renderEnumDefinition = function (e, enumName) {
            var codableString = _this.getCodableString();
            _this.emitBlock(["enum ", enumName, codableString], function () {
                _this.forEachCase(e, "none", function (name) {
                    _this.emitLine("case ", name);
                });
            });
        };
        _this.renderUnionDefinition = function (u, unionName) {
            var _a = Type_1.removeNullFromUnion(u), maybeNull = _a[0], nonNulls = _a[1];
            var codableString = _this.getCodableString();
            _this.emitBlock(["enum ", unionName, codableString], function () {
                nonNulls.forEach(function (t) {
                    _this.emitLine("case ", _this.unionFieldName(t), "(", _this.swiftType(t), ")");
                });
                if (maybeNull) {
                    _this.emitLine("case ", _this.unionFieldName(maybeNull));
                }
            });
        };
        _this.renderTopLevelExtensions4 = function (t, topLevelName) {
            var typeSource = _this.swiftType(t);
            var extensionSource;
            if (t instanceof Type_1.ArrayType) {
                extensionSource = ["Array where Element == ", _this.swiftType(t.items)];
            }
            else if (t instanceof Type_1.MapType) {
                extensionSource = ["Dictionary where Key == String, Value == ", _this.swiftType(t.values)];
            }
            else {
                extensionSource = typeSource;
            }
            _this.emitBlock(["extension ", extensionSource], function () {
                _this.emitBlock(["static func from(json: String, using encoding: String.Encoding = .utf8) -> ", typeSource, "?"], function () {
                    _this.emitLine("guard let data = json.data(using: encoding) else { return nil }");
                    _this.emitLine("return from(data: data)");
                });
                _this.emitNewline();
                _this.emitBlock(["static func from(data: Data) -> ", typeSource, "?"], function () {
                    _this.emitLine("let decoder = JSONDecoder()");
                    _this.emitLine("return try? decoder.decode(", typeSource, ".self, from: data)");
                });
                _this.emitNewline();
                _this.emitBlock(["static func from(url urlString: String) -> ", typeSource, "?"], function () {
                    _this.emitLine("guard let url = URL(string: urlString) else { return nil }");
                    _this.emitLine("guard let data = try? Data(contentsOf: url) else { return nil }");
                    _this.emitLine("return from(data: data)");
                });
                _this.emitNewline();
                _this.emitBlock("var jsonData: Data?", function () {
                    _this.emitLine("let encoder = JSONEncoder()");
                    _this.emitLine("return try? encoder.encode(self)");
                });
                _this.emitNewline();
                _this.emitBlock("var jsonString: String?", function () {
                    _this.emitLine("guard let data = self.jsonData else { return nil }");
                    _this.emitLine("return String(data: data, encoding: .utf8)");
                });
            });
        };
        _this.renderClassExtensions4 = function (c, className) {
            if (c.properties.isEmpty())
                return;
            _this.emitBlock(["extension ", className], function () {
                _this.emitBlock("enum CodingKeys: String, CodingKey", function () {
                    _this.forEachProperty(c, "none", function (name, jsonName) {
                        _this.emitLine("case ", name, ' = "', stringEscape(jsonName), '"');
                    });
                });
            });
        };
        _this.emitDecodingError = function (name) {
            _this.emitLine("throw DecodingError.typeMismatch(", name, '.self, DecodingError.Context(codingPath: decoder.codingPath, debugDescription: "Wrong type for ', name, '"))');
        };
        _this.renderEnumExtensions4 = function (e, enumName) {
            _this.emitBlock(["extension ", enumName], function () {
                _this.emitBlock("init(from decoder: Decoder) throws", function () {
                    _this.emitLine("let container = try decoder.singleValueContainer()");
                    _this.emitBlock(["if let x = try? container.decode(String.self)"], function () {
                        _this.emitLine("switch x {");
                        _this.forEachCase(e, "none", function (name, jsonName) {
                            _this.emitLine('case "', stringEscape(jsonName), '":');
                            _this.indent(function () {
                                _this.emitLine("self = .", name);
                                _this.emitLine("return");
                            });
                        });
                        _this.emitLine("default:");
                        _this.indent(function () { return _this.emitLine("break"); });
                        _this.emitLine("}");
                    });
                    _this.emitDecodingError(enumName);
                });
                _this.emitNewline();
                _this.emitBlock("func encode(to encoder: Encoder) throws", function () {
                    _this.emitLine("var container = encoder.singleValueContainer()");
                    _this.emitLine("switch self {");
                    _this.forEachCase(e, "none", function (name, jsonName) {
                        _this.emitLine("case .", name, ":");
                        _this.indent(function () { return _this.emitLine('try container.encode("', stringEscape(jsonName), '")'); });
                    });
                    _this.emitLine("}");
                });
            });
        };
        _this.renderUnionCase = function (t) {
            _this.emitBlock(["if let x = try? container.decode(", _this.swiftType(t), ".self)"], function () {
                _this.emitLine("self = .", _this.unionFieldName(t), "(x)");
                _this.emitLine("return");
            });
        };
        _this.renderUnionExtensions4 = function (u, unionName) {
            var _a = Type_1.removeNullFromUnion(u), maybeNull = _a[0], nonNulls = _a[1];
            _this.emitBlock(["extension ", unionName], function () {
                _this.emitBlock("init(from decoder: Decoder) throws", function () {
                    _this.emitLine("let container = try decoder.singleValueContainer()");
                    var boolMember = u.findMember("bool");
                    if (boolMember)
                        _this.renderUnionCase(boolMember);
                    var integerMember = u.findMember("integer");
                    if (integerMember)
                        _this.renderUnionCase(integerMember);
                    nonNulls.forEach(function (t) {
                        if (t.kind === "bool" || t.kind === "integer")
                            return;
                        _this.renderUnionCase(t);
                    });
                    if (maybeNull) {
                        _this.emitBlock("if container.decodeNil()", function () {
                            _this.emitLine("self = .", _this.unionFieldName(maybeNull));
                            _this.emitLine("return");
                        });
                    }
                    _this.emitDecodingError(unionName);
                });
                _this.emitNewline();
                _this.emitBlock("func encode(to encoder: Encoder) throws", function () {
                    _this.emitLine("var container = encoder.singleValueContainer()");
                    _this.emitLine("switch self {");
                    nonNulls.forEach(function (t) {
                        _this.emitLine("case .", _this.unionFieldName(t), "(let x):");
                        _this.indent(function () { return _this.emitLine("try container.encode(x)"); });
                    });
                    if (maybeNull) {
                        _this.emitLine("case .", _this.unionFieldName(maybeNull), ":");
                        _this.indent(function () { return _this.emitLine("try container.encodeNil()"); });
                    }
                    _this.emitLine("}");
                });
            });
        };
        _this.emitSupportFunctions4 = function () {
            var anyAndNullSet = Type_1.filterTypes(function (t) { return t.kind === "any" || t.kind === "null"; }, _this.topLevels);
            var needAny = anyAndNullSet.some(function (t) { return t.kind === "any"; });
            var needNull = anyAndNullSet.some(function (t) { return t.kind === "null"; });
            if (needAny || needNull) {
                _this.emitLine("// Helpers");
                _this.emitNewline();
                _this.emitMultiline("class JSONNull: Codable {\n    public init() {\n    }\n    \n    public required init(from decoder: Decoder) throws {\n        let container = try decoder.singleValueContainer()\n        if !container.decodeNil() {\n            throw DecodingError.typeMismatch(JSONNull.self, DecodingError.Context(codingPath: decoder.codingPath, debugDescription: \"Wrong type for JSONNull\"))\n        }\n    }\n    \n    public func encode(to encoder: Encoder) throws {\n        var container = encoder.singleValueContainer()\n        try container.encodeNil()\n    }\n}");
            }
            if (needAny) {
                _this.emitNewline();
                _this.emitMultiline("class JSONCodingKey : CodingKey {\n    let key: String\n    \n    required init?(intValue: Int) {\n        return nil\n    }\n    \n    required init?(stringValue: String) {\n        key = stringValue\n    }\n    \n    var intValue: Int? {\n        return nil\n    }\n    \n    var stringValue: String {\n        return key\n    }\n}\n\nclass JSONAny: Codable {\n    public let value: Any\n    \n    static func decodingError(forCodingPath codingPath: [CodingKey]) -> DecodingError {\n        let context = DecodingError.Context(codingPath: codingPath, debugDescription: \"Cannot decode JSONAny\")\n        return DecodingError.typeMismatch(JSONAny.self, context)\n    }\n    \n    static func encodingError(forValue value: Any, codingPath: [CodingKey]) -> EncodingError {\n        let context = EncodingError.Context(codingPath: codingPath, debugDescription: \"Cannot encode JSONAny\")\n        return EncodingError.invalidValue(value, context)\n    }\n\n    static func decode(from container: SingleValueDecodingContainer) throws -> Any {\n        if let value = try? container.decode(Bool.self) {\n            return value\n        }\n        if let value = try? container.decode(Int64.self) {\n            return value\n        }\n        if let value = try? container.decode(Double.self) {\n            return value\n        }\n        if let value = try? container.decode(String.self) {\n            return value\n        }\n        if container.decodeNil() {\n            return JSONNull()\n        }\n        throw decodingError(forCodingPath: container.codingPath)\n    }\n    \n    static func decode(from container: inout UnkeyedDecodingContainer) throws -> Any {\n        if let value = try? container.decode(Bool.self) {\n            return value\n        }\n        if let value = try? container.decode(Int64.self) {\n            return value\n        }\n        if let value = try? container.decode(Double.self) {\n            return value\n        }\n        if let value = try? container.decode(String.self) {\n            return value\n        }\n        if let value = try? container.decodeNil() {\n            if value {\n                return JSONNull()\n            }\n        }\n        if var container = try? container.nestedUnkeyedContainer() {\n            return try decodeArray(from: &container)\n        }\n        if var container = try? container.nestedContainer(keyedBy: JSONCodingKey.self) {\n            return try decodeDictionary(from: &container)\n        }\n        throw decodingError(forCodingPath: container.codingPath)\n    }\n    \n    static func decode(from container: inout KeyedDecodingContainer<JSONCodingKey>, forKey key: JSONCodingKey) throws -> Any {\n        if let value = try? container.decode(Bool.self, forKey: key) {\n            return value\n        }\n        if let value = try? container.decode(Int64.self, forKey: key) {\n            return value\n        }\n        if let value = try? container.decode(Double.self, forKey: key) {\n            return value\n        }\n        if let value = try? container.decode(String.self, forKey: key) {\n            return value\n        }\n        if let value = try? container.decodeNil(forKey: key) {\n            if value {\n                return JSONNull()\n            }\n        }\n        if var container = try? container.nestedUnkeyedContainer(forKey: key) {\n            return try decodeArray(from: &container)\n        }\n        if var container = try? container.nestedContainer(keyedBy: JSONCodingKey.self, forKey: key) {\n            return try decodeDictionary(from: &container)\n        }\n        throw decodingError(forCodingPath: container.codingPath)\n    }\n    \n    static func decodeArray(from container: inout UnkeyedDecodingContainer) throws -> [Any] {\n        var arr: [Any] = []\n        while !container.isAtEnd {\n            let value = try decode(from: &container)\n            arr.append(value)\n        }\n        return arr\n    }\n\n    static func decodeDictionary(from container: inout KeyedDecodingContainer<JSONCodingKey>) throws -> [String: Any] {\n        var dict = [String: Any]()\n        for key in container.allKeys {\n            let value = try decode(from: &container, forKey: key)\n            dict[key.stringValue] = value\n        }\n        return dict\n    }\n    \n    static func encode(to container: inout UnkeyedEncodingContainer, array: [Any]) throws {\n        for value in array {\n            if let value = value as? Bool {\n                try container.encode(value)\n            } else if let value = value as? Int64 {\n                try container.encode(value)\n            } else if let value = value as? Double {\n                try container.encode(value)\n            } else if let value = value as? String {\n                try container.encode(value)\n            } else if value is JSONNull {\n                try container.encodeNil()\n            } else if let value = value as? [Any] {\n                var container = container.nestedUnkeyedContainer()\n                try encode(to: &container, array: value)\n            } else if let value = value as? [String: Any] {\n                var container = container.nestedContainer(keyedBy: JSONCodingKey.self)\n                try encode(to: &container, dictionary: value)\n            } else {\n                throw encodingError(forValue: value, codingPath: container.codingPath)\n            }\n        }\n    }\n    \n    static func encode(to container: inout KeyedEncodingContainer<JSONCodingKey>, dictionary: [String: Any]) throws {\n        for (key, value) in dictionary {\n            let key = JSONCodingKey(stringValue: key)!\n            if let value = value as? Bool {\n                try container.encode(value, forKey: key)\n            } else if let value = value as? Int64 {\n                try container.encode(value, forKey: key)\n            } else if let value = value as? Double {\n                try container.encode(value, forKey: key)\n            } else if let value = value as? String {\n                try container.encode(value, forKey: key)\n            } else if value is JSONNull {\n                try container.encodeNil(forKey: key)\n            } else if let value = value as? [Any] {\n                var container = container.nestedUnkeyedContainer(forKey: key)\n                try encode(to: &container, array: value)\n            } else if let value = value as? [String: Any] {\n                var container = container.nestedContainer(keyedBy: JSONCodingKey.self, forKey: key)\n                try encode(to: &container, dictionary: value)\n            } else {\n                throw encodingError(forValue: value, codingPath: container.codingPath)\n            }\n        }\n    }\n\n    static func encode(to container: inout SingleValueEncodingContainer, value: Any) throws {\n        if let value = value as? Bool {\n            try container.encode(value)\n        } else if let value = value as? Int64 {\n            try container.encode(value)\n        } else if let value = value as? Double {\n            try container.encode(value)\n        } else if let value = value as? String {\n            try container.encode(value)\n        } else if value is JSONNull {\n            try container.encodeNil()\n        } else {\n            throw encodingError(forValue: value, codingPath: container.codingPath)\n        }\n    }\n    \n    public required init(from decoder: Decoder) throws {\n        if var arrayContainer = try? decoder.unkeyedContainer() {\n            self.value = try JSONAny.decodeArray(from: &arrayContainer)\n        } else if var container = try? decoder.container(keyedBy: JSONCodingKey.self) {\n            self.value = try JSONAny.decodeDictionary(from: &container)\n        } else {\n            let container = try decoder.singleValueContainer()\n            self.value = try JSONAny.decode(from: container)\n        }\n    }\n    \n    public func encode(to encoder: Encoder) throws {\n        if let arr = self.value as? [Any] {\n            var container = encoder.unkeyedContainer()\n            try JSONAny.encode(to: &container, array: arr)\n        } else if let dict = self.value as? [String: Any] {\n            var container = encoder.container(keyedBy: JSONCodingKey.self)\n            try JSONAny.encode(to: &container, dictionary: dict)\n        } else {\n            var container = encoder.singleValueContainer()\n            try JSONAny.encode(to: &container, value: self.value)\n        }\n    }\n}");
            }
        };
        _this.emitMark = function (line, horizontalLine) {
            if (horizontalLine === void 0) { horizontalLine = false; }
            _this.emitLine("// MARK: ", line, horizontalLine ? " -" : "");
        };
        return _this;
    }
    Object.defineProperty(SwiftRenderer.prototype, "forbiddenNamesForGlobalNamespace", {
        get: function () {
            return keywords;
        },
        enumerable: true,
        configurable: true
    });
    SwiftRenderer.prototype.forbiddenForProperties = function (c, classNamed) {
        return { names: [], namespaces: [this.globalNamespace] };
    };
    SwiftRenderer.prototype.forbiddenForCases = function (e, enumNamed) {
        return { names: [], namespaces: [this.globalNamespace] };
    };
    SwiftRenderer.prototype.topLevelNameStyle = function (rawName) {
        return swiftNameStyle(true, rawName);
    };
    Object.defineProperty(SwiftRenderer.prototype, "namedTypeNamer", {
        get: function () {
            return upperNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SwiftRenderer.prototype, "propertyNamer", {
        get: function () {
            return lowerNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SwiftRenderer.prototype, "caseNamer", {
        get: function () {
            return lowerNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    SwiftRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        if (type.isNamedType()) {
            return type;
        }
        return null;
    };
    SwiftRenderer.prototype.emitSourceStructure = function () {
        var _this = this;
        this.renderHeader();
        this.forEachTopLevel("leading", this.renderTopLevelAlias, function (t) { return !_this.namedTypeToNameForTopLevel(t); });
        this.forEachNamedType("leading-and-interposing", false, this.renderClassDefinition, this.renderEnumDefinition, this.renderUnionDefinition);
        if (!this._justTypes) {
            this.emitNewline();
            this.emitMark("Top-level extensions", true);
            this.forEachTopLevel("leading-and-interposing", this.renderTopLevelExtensions4);
            this.emitNewline();
            this.emitMark("Codable extensions", true);
            this.forEachNamedType("leading-and-interposing", false, this.renderClassExtensions4, this.renderEnumExtensions4, this.renderUnionExtensions4);
            this.emitNewline();
            this.emitSupportFunctions4();
        }
    };
    return SwiftRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
