"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var unicode = require("unicode-properties");
function computeAsciiMap(mapper) {
    var charStringMap = [];
    var charNoEscapeMap = [];
    for (var i = 0; i < 128; i++) {
        var noEscape = 0;
        var result = mapper(i);
        if (result === String.fromCharCode(i)) {
            noEscape = 1;
        }
        charStringMap.push(result);
        charNoEscapeMap.push(noEscape);
    }
    return { charStringMap: charStringMap, charNoEscapeMap: charNoEscapeMap };
}
// FIXME: This is a copy of code in src/Data/String/Util.js
function utf16ConcatMap(mapper) {
    var _a = computeAsciiMap(mapper), charStringMap = _a.charStringMap, charNoEscapeMap = _a.charNoEscapeMap;
    return function stringConcatMap_inner(s) {
        var cs = null;
        var start = 0;
        var i = 0;
        while (i < s.length) {
            var cc = s.charCodeAt(i);
            if (!charNoEscapeMap[cc]) {
                if (cs === null)
                    cs = [];
                cs.push(s.substring(start, i));
                var str = charStringMap[cc];
                if (str === undefined) {
                    cs.push(mapper(s.charCodeAt(i)));
                }
                else {
                    cs.push(str);
                }
                start = i + 1;
            }
            i++;
        }
        if (cs === null)
            return s;
        cs.push(s.substring(start, i));
        return cs.join("");
    };
}
exports.utf16ConcatMap = utf16ConcatMap;
function utf32ConcatMap(mapper) {
    var _a = computeAsciiMap(mapper), charStringMap = _a.charStringMap, charNoEscapeMap = _a.charNoEscapeMap;
    return function stringConcatMap_inner(s) {
        var cs = null;
        var start = 0;
        var i = 0;
        while (i < s.length) {
            var cc = s.charCodeAt(i);
            if (!charNoEscapeMap[cc]) {
                if (cs === null)
                    cs = [];
                cs.push(s.substring(start, i));
                if (cc >= 0xd800 && cc <= 0xdbff) {
                    var highSurrogate = cc;
                    i++;
                    var lowSurrogate = s.charCodeAt(i);
                    if (lowSurrogate < 0xdc00 || lowSurrogate > 0xdfff) {
                        throw "High surrogate not followed by low surrogate";
                    }
                    var highBits = highSurrogate - 0xd800;
                    var lowBits = lowSurrogate - 0xdc00;
                    cc = 0x10000 + lowBits + (highBits << 10);
                }
                var str = charStringMap[cc];
                if (str === undefined) {
                    cs.push(mapper(cc));
                }
                else {
                    cs.push(str);
                }
                start = i + 1;
            }
            i++;
        }
        if (cs === null)
            return s;
        cs.push(s.substring(start, i));
        return cs.join("");
    };
}
exports.utf32ConcatMap = utf32ConcatMap;
function utf16LegalizeCharacters(isLegal) {
    return utf16ConcatMap(function (u) { return (isLegal(u) ? String.fromCharCode(u) : "_"); });
}
exports.utf16LegalizeCharacters = utf16LegalizeCharacters;
function legalizeCharacters(isLegal) {
    return utf32ConcatMap(function (u) { return (u <= 0xffff && isLegal(u) ? String.fromCharCode(u) : "_"); });
}
exports.legalizeCharacters = legalizeCharacters;
function intToHex(i, width) {
    var str = i.toString(16);
    if (str.length >= width)
        return str;
    return "0".repeat(width - str.length) + str;
}
exports.intToHex = intToHex;
function standardUnicodeHexEscape(codePoint) {
    if (codePoint <= 0xffff) {
        return "\\u" + intToHex(codePoint, 4);
    }
    else {
        return "\\U" + intToHex(codePoint, 8);
    }
}
exports.standardUnicodeHexEscape = standardUnicodeHexEscape;
function escapeNonPrintableMapper(printablePredicate, escaper) {
    function mapper(u) {
        switch (u) {
            case 0x5c:
                return "\\\\";
            case 0x22:
                return '\\"';
            case 0x0a:
                return "\\n";
            case 0x09:
                return "\\t";
            default:
                if (printablePredicate(u)) {
                    return String.fromCharCode(u);
                }
                return escaper(u);
        }
    }
    return mapper;
}
exports.escapeNonPrintableMapper = escapeNonPrintableMapper;
exports.utf16StringEscape = utf16ConcatMap(escapeNonPrintableMapper(isPrintable, standardUnicodeHexEscape));
exports.stringEscape = utf32ConcatMap(escapeNonPrintableMapper(isPrintable, standardUnicodeHexEscape));
function isPrintable(codePoint) {
    if (codePoint > 0xffff)
        return false;
    var category = unicode.getCategory(codePoint);
    return ([
        "Mc",
        "No",
        "Sk",
        "Me",
        "Nd",
        "Po",
        "Lt",
        "Pc",
        "Sm",
        "Zs",
        "Lu",
        "Pd",
        "So",
        "Pe",
        "Pf",
        "Ps",
        "Sc",
        "Ll",
        "Lm",
        "Pi",
        "Nl",
        "Mn",
        "Lo"
    ].indexOf(category) >= 0);
}
exports.isPrintable = isPrintable;
function isAscii(codePoint) {
    return codePoint < 128;
}
exports.isAscii = isAscii;
function isLetter(codePoint) {
    var category = unicode.getCategory(codePoint);
    // FIXME: Include Letter, modifier (Lm)?
    return ["Lu", "Ll", "Lt", "Lo"].indexOf(category) >= 0;
}
exports.isLetter = isLetter;
function isDigit(codePoint) {
    var category = unicode.getCategory(codePoint);
    return ["Nd"].indexOf(category) >= 0;
}
exports.isDigit = isDigit;
function isNumeric(codePoint) {
    var category = unicode.getCategory(codePoint);
    return ["No", "Nd", "Nl"].indexOf(category) >= 0;
}
exports.isNumeric = isNumeric;
function isLetterOrUnderscore(codePoint) {
    return isLetter(codePoint) || codePoint === 0x5f;
}
exports.isLetterOrUnderscore = isLetterOrUnderscore;
function isLetterOrUnderscoreOrDigit(codePoint) {
    return isLetterOrUnderscore(codePoint) || isDigit(codePoint);
}
exports.isLetterOrUnderscoreOrDigit = isLetterOrUnderscoreOrDigit;
function modifyFirstChar(f, s) {
    if (s === "")
        return s;
    return f(s[0]) + s.slice(1);
}
function capitalize(str) {
    return modifyFirstChar(function (c) { return c.toUpperCase(); }, str);
}
exports.capitalize = capitalize;
function decapitalize(str) {
    return modifyFirstChar(function (c) { return c.toLowerCase(); }, str);
}
exports.decapitalize = decapitalize;
var wordSeparatorRegex = /[-_. ]+/;
function pascalCase(str) {
    var words = str.split(wordSeparatorRegex).map(capitalize);
    return words.join("");
}
exports.pascalCase = pascalCase;
function camelCase(str) {
    return decapitalize(pascalCase(str));
}
exports.camelCase = camelCase;
function snakeCase(str) {
    var separated = str.replace(/([^A-Z])([A-Z])/g, "$1_$2");
    var words = separated.split(wordSeparatorRegex).map(decapitalize);
    return words.join("_");
}
exports.snakeCase = snakeCase;
function upperUnderscoreCase(str) {
    var separated = str.replace(/([^A-Z])([A-Z])/g, "$1_$2");
    var words = separated.split(wordSeparatorRegex).map(function (s) { return s.toUpperCase(); });
    return words.join("_");
}
exports.upperUnderscoreCase = upperUnderscoreCase;
function startWithLetter(isAllowedStart, // FIXME: technically, this operates on UTF16 units
    upper, str) {
    var modify = upper ? capitalize : decapitalize;
    if (str === "")
        return modify("empty");
    if (isAllowedStart(str.charCodeAt(0)))
        return modify(str);
    return modify("the" + str);
}
exports.startWithLetter = startWithLetter;
function intercalate(separator, items) {
    var acc = [];
    items.forEach(function (x) {
        if (acc.length > 0)
            acc.push(separator);
        acc.push(x);
    });
    return immutable_1.List(acc);
}
exports.intercalate = intercalate;
function defined(x) {
    if (x !== undefined)
        return x;
    throw "Defined value expected, but got undefined";
}
exports.defined = defined;
function nonNull(x) {
    if (x !== null)
        return x;
    throw "Non-null value expected, but got null";
}
exports.nonNull = nonNull;
function assertNever(x) {
    throw new Error("Unexpected object: " + x);
}
exports.assertNever = assertNever;
